from typing import Never
import lucid
from lucid._tensor import Tensor


def _prob_check(p: float) -> Never:
    if not 0 <= p < 1:
        raise ValueError("Dropout probability `p` must be in the range [0, 1).")


def dropout(input_: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    _prob_check(p)
    if not training:
        return input_

    mask = lucid.random.rand(*input_.shape) > p
    scale = 1.0 / (1 - p)
    return input_ * mask * scale


def dropoutnd(input_: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    _prob_check(p)
    if not training:
        return input_

    spatial_dim = input_.ndim - 2
    mask = lucid.random.rand(*input_.shape[:2], *(1,) * spatial_dim) > p
    scale = 1.0 / (1 - p)
    return input_ * mask * scale


def alpha_dropout(input_: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    _prob_check(p)
    if not training:
        return input_

    _alpha = -1.7580993408473766
    _lambda = 1.0507009873554805

    mask = lucid.random.rand(*input_.shape) > p
    scale = 1.0 / (1 - p)

    dropped = input_ * mask * scale
    noise = (1 - mask) * _alpha * _lambda
    return dropped + noise



import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor


__all__ = ["ReLU", "LeakyReLU", "ELU", "SELU", "GELU", "Sigmoid", "Tanh"]


class ReLU(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return F.relu(input_)


class LeakyReLU(nn.Module):
    def __init__(self, negative_slope: float = 0.01) -> None:
        super().__init__()
        self.negative_slope = negative_slope

    def forward(self, input_: Tensor) -> Tensor:
        return F.leaky_relu(input_, self.negative_slope)


class ELU(nn.Module):
    def __init__(self, alpha: float = 1.0) -> None:
        super().__init__()
        self.alpha = alpha

    def forward(self, input_: Tensor) -> Tensor:
        return F.elu(input_, self.alpha)


class SELU(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return F.selu(input_)


class GELU(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return F.gelu(input_)


class Sigmoid(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return F.sigmoid(input_)


class Tanh(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return F.tanh(input_)



from lucid._tensor import Tensor
from lucid.linalg import _func


def inv(a: Tensor) -> Tensor:
    return _func.inv(a)


def det(a: Tensor) -> Tensor:
    return _func.det(a)


def solve(a: Tensor, b: Tensor) -> Tensor:
    return _func.solve(a, b)


def cholesky(a: Tensor) -> Tensor:
    return _func.cholesky(a)


def norm(a: Tensor, ord: int = 2) -> Tensor:
    return _func.norm(a, ord)


def eig(a: Tensor) -> tuple[Tensor, Tensor]:
    return _func.eig(a)


def qr(a: Tensor) -> tuple[Tensor, Tensor]:
    return _func.qr(a)


def svd(a: Tensor, full_matrices: bool = True) -> tuple[Tensor, Tensor, Tensor]:
    return _func.svd(a, full_matrices)


def matrix_power(a: Tensor, n: int) -> Tensor:
    return _func.matrix_power(a, n)


def pinv(a: Tensor) -> Tensor:
    return _func.pinv(a)



from typing import Literal

import lucid
from lucid._tensor import Tensor

_ReductionType = Literal["mean", "sum"]


def _loss_reduction(loss: Tensor, reduction: _ReductionType | None) -> Tensor:
    match reduction:
        case "mean":
            return loss.mean()
        case "sum":
            return loss.sum()
        case None:
            return loss
        case _:
            raise ValueError(
                "Invalid reduction type. Choose 'mean', 'sum', or 'none'.",
            )


def mse_loss(
    input_: Tensor, target: Tensor, reduction: _ReductionType | None = "mean"
) -> Tensor:
    loss = (input_ - target) ** 2
    return _loss_reduction(loss, reduction)


def binary_cross_entropy(
    input_: Tensor,
    target: Tensor,
    weight: Tensor | None = None,
    reduction: _ReductionType | None = "mean",
    eps: float = 1e-7,
) -> Tensor:
    input_ = lucid.clip(input_, eps, 1 - eps)
    loss = -target * lucid.log(input_) - (1 - target) * lucid.log(1 - input_)

    if weight is not None:
        loss *= weight

    return _loss_reduction(loss, reduction)


def cross_entropy(
    input_: Tensor,
    target: Tensor,
    weight: Tensor | None = None,
    reduction: _ReductionType | None = "mean",
    eps: float = 1e-7,
) -> Tensor:
    exp_logits = lucid.exp(input_ - lucid.max(input_, axis=1, keepdims=True))
    prob = exp_logits / lucid.sum(exp_logits, axis=1, keepdims=True)

    loss = -lucid.log(prob[:, target.data.astype(int)] + eps)
    if weight is not None:
        loss *= weight[target.data.astype(int)]

    return _loss_reduction(loss, reduction)


def nll_loss(
    input_: Tensor,
    target: Tensor,
    weight: Tensor | None = None,
    reduction: _ReductionType | None = "mean",
) -> Tensor:
    loss = -input_[:, target.data.astype(int)]
    if weight is not None:
        loss *= weight[target.data.astype(int)]

    return _loss_reduction(loss, reduction)


def huber_loss(
    input_: Tensor,
    target: Tensor,
    delta: float = 1.0,
    reduction: _ReductionType | None = "mean",
) -> Tensor:
    diff = lucid.abs(input_ - target)
    quad = lucid.minimum(diff, delta)
    linear = diff - quad
    loss = 0.5 * quad**2 + delta * linear

    return _loss_reduction(loss, reduction)



import numpy as np

from lucid._backend import (
    create_func_op,
    create_bfunc_op,
    create_ufunc_op,
    _FuncOpReturnType,
)
from lucid._tensor import Tensor
from lucid.types import _ArrayOrScalar


@create_ufunc_op()
def inv(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.linalg.inv(self.data))

    def compute_grad() -> _ArrayOrScalar:
        return -np.dot(np.dot(result.data.T, result.grad), result.data)

    return result, compute_grad


@create_ufunc_op()
def det(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.linalg.det(self.data))

    def compute_grad() -> _ArrayOrScalar:
        return result.data * np.linalg.inv(self.data).T * result.grad

    return result, compute_grad


@create_bfunc_op()
def solve(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.linalg.solve(self.data, other.data))

    def compute_grad() -> tuple[_ArrayOrScalar, _ArrayOrScalar]:
        inv_A = np.linalg.inv(self.data)
        self_grad = -inv_A @ (result.grad @ result.data.T) @ inv_A
        other_grad = inv_A @ result.grad

        return self_grad, other_grad

    return result, compute_grad


@create_ufunc_op()
def cholesky(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.linalg.cholesky(self.data))

    def compute_grad() -> _ArrayOrScalar:
        return 2 * result.data * result.grad

    return result, compute_grad


@create_ufunc_op()
def norm(self: Tensor, ord: int = 2) -> _FuncOpReturnType:
    if not isinstance(ord, int):
        raise NotImplementedError("Only integer p-norms are supported.")

    result = Tensor(np.linalg.norm(self.data, ord=ord))

    def compute_grad() -> _ArrayOrScalar:
        if ord == 2:
            grad = (
                self.data / result.data
                if result.data != 0
                else np.zeros_like(self.data)
            )
        elif ord == 1:
            grad = np.sign(self.data)
        else:
            grad = (
                (np.abs(self.data) ** (ord - 1))
                * np.sign(self.data)
                / (result.data ** (ord - 1))
                if result.data != 0
                else np.zeros_like(self.data)
            )

        return grad * result.grad

    return result, compute_grad


@create_func_op(n_in=1, n_ret=2)
def eig(self: Tensor) -> _FuncOpReturnType:
    eigvals, eigvecs = np.linalg.eig(self.data)
    ndim = self.shape[-2]

    result_eigvals = Tensor(eigvals)
    result_eigvecs = Tensor(eigvecs / np.linalg.norm(eigvecs, axis=-2, keepdims=True))

    def compute_grad_eigvals() -> _ArrayOrScalar:
        grad = np.einsum(
            "...k,...ki,...kj->...ij", result_eigvals.grad, eigvecs, eigvecs
        )
        return grad

    def compute_grad_eigvecs(_eps: float = 1e-12) -> _ArrayOrScalar:
        eigval_diffs = eigvals[..., :, np.newaxis] - eigvals[..., np.newaxis, :]
        eigval_diffs += np.eye(ndim)[..., :, :] * _eps

        inv_eigval_diffs = 1.0 / eigval_diffs
        for index in np.ndindex(inv_eigval_diffs.shape[:-2]):
            np.fill_diagonal(inv_eigval_diffs[index], 0.0)

        outer_prods = np.einsum("...ip,...jq->...pqij", eigvecs, eigvecs)
        S = np.einsum("...kp,...pqij->...pij", inv_eigval_diffs, outer_prods)

        grad = np.einsum("...pk,...pij,...ki->...ij", result_eigvecs.grad, S, eigvecs)
        return grad

    return (
        (result_eigvals, compute_grad_eigvals),
        (result_eigvecs, compute_grad_eigvecs),
    )


@create_func_op(n_in=1, n_ret=2)
def qr(self: Tensor) -> _FuncOpReturnType:
    Q, R = np.linalg.qr(self.data)

    result_q = Tensor(Q)
    result_r = Tensor(R)

    def compute_grad_q() -> _ArrayOrScalar:
        grad_q = result_q.grad
        qt_grad_q = np.einsum("...ik,...kj->...ij", Q.mT, grad_q)
        qt_grad_q_r = np.einsum("...ij,...jk->...ik", qt_grad_q, R)

        return np.einsum("...ij,...jk->...ik", grad_q, R) - np.einsum(
            "...ij,...jk->...ik", Q, qt_grad_q_r
        )

    def compute_grad_r() -> _ArrayOrScalar:
        grad_r = result_r.grad
        return np.einsum("...ij,...jk->...ik", Q, grad_r)

    return (result_q, compute_grad_q), (result_r, compute_grad_r)


@create_func_op(n_in=1, n_ret=3)
def svd(self: Tensor, full_matrices: bool = True) -> _FuncOpReturnType:
    U, S, VT = np.linalg.svd(self.data, full_matrices=full_matrices)

    result_u = Tensor(U)
    result_s = Tensor(S)
    result_vt = Tensor(VT)

    def compute_grad_u() -> _ArrayOrScalar:
        return np.einsum("...ik,...k,...jk->...ij", result_u.grad, S, VT.mT)

    def compute_grad_s() -> _ArrayOrScalar:
        return np.einsum("...ik,...k,...jk->...ij", U, result_s.grad, VT.mT)

    def compute_grad_vt() -> _ArrayOrScalar:
        return np.einsum("...ik,...k,...jk->...ij", U, S, result_vt.grad.mT)

    return (
        (result_u, compute_grad_u),
        (result_s, compute_grad_s),
        (result_vt, compute_grad_vt),
    )


@create_ufunc_op()
def matrix_power(self: Tensor, n: int) -> _FuncOpReturnType:
    result = Tensor(np.linalg.matrix_power(self.data, n))

    def compute_grad() -> _ArrayOrScalar:
        grad = np.zeros_like(self.data)
        if n == 0:
            return grad
        else:
            for k in range(abs(n)):
                left_exp = n - np.sign(n) * k - np.sign(n)
                right_exp = np.sign(n) * k

                left = np.linalg.matrix_power(self.data, left_exp)
                right = np.linalg.matrix_power(self.data, right_exp)

                grad += left @ result.grad @ right
            if n < 0:
                grad = -grad

        return grad

    return result, compute_grad


@create_ufunc_op()
def pinv(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.linalg.pinv(self.data))

    def compute_grad() -> _ArrayOrScalar:
        U, S, Vh = np.linalg.svd(self.data, full_matrices=False)
        S_inv_squared = np.diag(1 / (S**2))

        term_1 = (
            Vh.T
            @ S_inv_squared
            @ U.T
            @ result.grad.T
            @ (np.eye(self.shape[0]) - self.data @ result.data)
        )
        term_2 = (
            (np.eye(self.shape[1]) - result.data @ self.data)
            @ result.grad.T
            @ U
            @ S_inv_squared
            @ Vh
        )
        grad = -result.data.T @ result.grad.T @ result.data.T + term_1 + term_2
        return grad.T

    return result, compute_grad



import numpy as np

from lucid._tensor import Tensor
from lucid._backend import create_bfunc_op, _FuncOpReturnType
from lucid.types import _NumPyArray


@create_bfunc_op()
def _add(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor(self.data + other.data)

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return result.grad, result.grad

    return result, compute_grad


@create_bfunc_op()
def _sub(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor(self.data - other.data)

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return result.grad, -result.grad

    return result, compute_grad


@create_bfunc_op()
def _mul(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor(self.data * other.data)

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return other.data * result.grad, self.data * result.grad

    return result, compute_grad


@create_bfunc_op()
def _truediv(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor(self.data / other.data)

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return (
            (1 / other.data) * result.grad,
            (-self.data / (other.data**2)) * result.grad,
        )

    return result, compute_grad


@create_bfunc_op(has_gradient=False)
def _equal(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor((self.data == other.data).astype(self.dtype))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return np.array(0.0), np.array(0.0)

    return result, compute_grad


@create_bfunc_op(has_gradient=False)
def _not_equal(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor((self.data != other.data).astype(self.dtype))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return np.array(0.0), np.array(0.0)

    return result, compute_grad


@create_bfunc_op(has_gradient=False)
def _greater(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor((self.data > other.data).astype(self.dtype))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return np.array(0.0), np.array(0.0)

    return result, compute_grad


@create_bfunc_op(has_gradient=False)
def _greater_or_equal(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor((self.data >= other.data).astype(self.dtype))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return np.array(0.0), np.array(0.0)

    return result, compute_grad


@create_bfunc_op(has_gradient=False)
def _less(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor((self.data < other.data).astype(self.dtype))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return np.array(0.0), np.array(0.0)

    return result, compute_grad


@create_bfunc_op(has_gradient=False)
def _less_or_equal(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor((self.data <= other.data).astype(self.dtype))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return np.array(0.0), np.array(0.0)

    return result, compute_grad


@create_bfunc_op()
def minimum(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.minimum(self.data, other.data))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        self_grad = (self.data <= other.data).astype(self.dtype)
        other_grad = (self.data > other.data).astype(other.dtype)
        return self_grad * result.grad, other_grad * result.grad

    return result, compute_grad


@create_bfunc_op()
def maximum(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.maximum(self.data, other.data))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        self_grad = (self.data >= other.data).astype(self.dtype)
        other_grad = (other.data > self.data).astype(other.dtype)
        return self_grad * result.grad, other_grad * result.grad

    return result, compute_grad


@create_bfunc_op()
def power(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.power(self.data, other.data))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        self_grad = other.data * np.power(self.data, other.data - 1)
        other_grad = np.power(self.data, other.data) * np.log(self.data)
        return self_grad * result.grad, other_grad * result.grad

    return result, compute_grad


@create_bfunc_op()
def dot(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.dot(self.data, other.data))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return (
            result.grad.dot(other.data.mT),
            self.data.mT.dot(result.grad),
        )

    return result, compute_grad


@create_bfunc_op()
def inner(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.inner(self.data, other.data))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return (
            np.tensordot(result.grad, other.data, axes=(-1, -1)),
            np.tensordot(self.data, result.grad, axes=(-1, -1)),
        )

    return result, compute_grad


@create_bfunc_op()
def outer(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.outer(self.data, other.data))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return (
            np.tensordot(result.grad, other.data, axes=(1, 0)),
            np.tensordot(result.grad, self.data, axes=(0, 0)),
        )

    return result, compute_grad


@create_bfunc_op()
def _matmul(self: Tensor, other: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.matmul(self.data, other.data))

    def compute_grad() -> tuple[_NumPyArray, _NumPyArray]:
        return (
            np.matmul(result.grad, other.data.mT),
            np.matmul(self.data.mT, result.grad),
        )

    return result, compute_grad


_radd: callable = lambda self, other: _add(self, other)
_rsub: callable = lambda self, other: _sub(other, self)
_rmul: callable = lambda self, other: _mul(self, other)
_rtruediv: callable = lambda self, other: _truediv(other, self)



from typing import Sequence
import numpy as np

from lucid._tensor import Tensor
from lucid.types import _ShapeLike, _NumPyArray, _ArrayLikeInt

from lucid._backend import create_ufunc_op, create_mfunc_op, _FuncOpReturnType


@create_ufunc_op()
def _reshape(self: Tensor, shape: _ShapeLike) -> _FuncOpReturnType:
    original_shape = self.shape
    result = Tensor(self.data.reshape(shape))

    def compute_grad() -> _NumPyArray:
        return result.grad.reshape(*original_shape)

    return result, compute_grad


@create_ufunc_op()
def _reshape_inplace(self: Tensor, *shape: int) -> _FuncOpReturnType:
    original_shape = self.shape
    result = Tensor(self.data.reshape(*shape))

    def compute_grad() -> _NumPyArray:
        return result.grad.reshape(*original_shape)

    return result, compute_grad


@create_ufunc_op()
def squeeze(self: Tensor, axis: _ShapeLike | None = None) -> _FuncOpReturnType:
    original_shape = self.shape
    result = Tensor(self.data.squeeze(axis=axis))

    def compute_grad() -> _NumPyArray:
        return result.grad.reshape(original_shape)

    return result, compute_grad


@create_ufunc_op()
def unsqueeze(self: Tensor, axis: _ShapeLike) -> _FuncOpReturnType:
    result = Tensor(np.expand_dims(self.data, axis=axis))

    def compute_grad() -> _NumPyArray:
        return result.grad.squeeze(axis=axis)

    return result, compute_grad


@create_ufunc_op()
def ravel(self: Tensor) -> _FuncOpReturnType:
    original_shape = self.shape
    result = Tensor(self.data.ravel())

    def compute_grad() -> _NumPyArray:
        return result.grad.reshape(original_shape)

    return result, compute_grad


@create_mfunc_op()
def stack(*tensors: Tensor, axis: int = 0) -> _FuncOpReturnType:
    data_arr = [tensor.data for tensor in tensors]
    result = Tensor(np.stack(data_arr, axis=axis))

    def compute_grad() -> tuple[_NumPyArray, ...]:
        split_grads = np.split(result.grad, len(tensors), axis=axis)
        return tuple(split_grads)

    return result, compute_grad


@create_mfunc_op()
def concatenate(*tensors: Tensor, axis: int = 0) -> _FuncOpReturnType:
    data_arr = [tensor.data for tensor in tensors]
    result = Tensor(np.concatenate(data_arr, axis=axis))

    def compute_grad() -> tuple[_NumPyArray, ...]:
        split_sizes = [tensor.shape[axis] for tensor in tensors]
        split_indices = np.cumsum(split_sizes)[:-1]
        split_grads = np.split(result.grad, split_indices, axis=axis)

        return tuple(split_grads)

    return result, compute_grad


@create_mfunc_op()
def hstack(*tensors: Tensor) -> _FuncOpReturnType:
    data_arr = [tensor.data for tensor in tensors]
    result = Tensor(np.hstack(data_arr))

    def compute_grad() -> tuple[_NumPyArray, ...]:
        split_sizes = [
            tensor.shape[1] if result.ndim > 1 else tensor.shape[0]
            for tensor in tensors
        ]
        split_indices = np.cumsum(split_sizes)[:-1]
        split_grads = (
            np.hsplit(result.grad, split_indices)
            if result.ndim > 1
            else np.split(result.grad, len(tensors))
        )

        return tuple(split_grads)

    return result, compute_grad


@create_mfunc_op()
def vstack(*tensors: Tensor) -> _FuncOpReturnType:
    data_arr = [tensor.data for tensor in tensors]
    result = Tensor(np.vstack(data_arr))

    def compute_grad() -> tuple[_NumPyArray, ...]:
        split_sizes = [tensor.shape[0] for tensor in tensors]
        split_indices = np.cumsum(split_sizes)[:-1]
        split_grads = np.split(result.grad, split_indices, axis=0)

        return tuple(split_grads)

    return result, compute_grad


@create_ufunc_op()
def pad(self: Tensor, pad_width: _ArrayLikeInt) -> _FuncOpReturnType:
    result = Tensor(np.pad(self.data, pad_width))

    def _normalize_pad_width(pad_width: _ArrayLikeInt, ndim: int) -> _ArrayLikeInt:
        if isinstance(pad_width, int):
            return ((pad_width, pad_width),) * ndim

        if isinstance(pad_width, (tuple, list)):
            pad_width = list(pad_width)
            if all(isinstance(pw, int) for pw in pad_width):
                if len(pad_width) == 1:
                    return ((pad_width[0], pad_width[0]),) * ndim
                elif len(pad_width) == 2:
                    return (tuple(pad_width),) * ndim
                elif len(pad_width) == ndim:
                    return tuple((pw, pw) for pw in pad_width)

            elif all(
                isinstance(pw, (tuple, list)) and len(pw) == 2 for pw in pad_width
            ):
                if len(pad_width) == ndim:
                    return tuple(tuple(pw) for pw in pad_width)
                elif len(pad_width) == 1:
                    return (tuple(pad_width[0]),) * ndim

        raise ValueError(f"Invalid pad_width format: '{pad_width}'.")

    pad_width_norm = _normalize_pad_width(pad_width, self.ndim)

    def compute_grad() -> _NumPyArray:
        grad_input = np.zeros_like(self.data)
        slices = []
        for pw in pad_width_norm:
            before, after = pw
            start = before
            end = -after if after != 0 else None
            slices.append(slice(start, end))

        grad_input = result.grad[tuple(slices)]
        return grad_input

    return result, compute_grad


@create_ufunc_op()
def repeat(
    self: Tensor, repeats: int | Sequence[int], axis: int | None = None
) -> _FuncOpReturnType:
    result = Tensor(np.repeat(self.data, repeats, axis=axis))

    def compute_grad() -> _NumPyArray:
        grad_input = np.zeros_like(self.data)
        repeats_arr = np.asarray(repeats)

        if axis is None:
            input_flat = self.data.flatten()
            grad_input_flat = grad_input.flatten()
            grad_output_flat = result.grad.flatten()

            input_size = input_flat.size

            if repeats_arr.size == 1:
                repeats_arr = np.full(input_size, repeats_arr)
            elif repeats_arr.size != input_size:
                raise ValueError(
                    "repeats must be an integer or a "
                    + "sequence of the same length as input."
                )

            input_indices = np.arange(input_size)
            output_indices = np.repeat(input_indices, repeats_arr)

            np.add.at(grad_input_flat, output_indices, grad_output_flat)
            grad_input = grad_input_flat.reshape(self.shape)

        else:
            if repeats_arr.size == 1:
                repeats_arr = np.full(self.shape[axis], repeats_arr)
            elif repeats_arr.size != self.shape[axis]:
                raise ValueError(
                    "repeats must be an integer or a "
                    + "sequence of the same length as the axis dimension."
                )

            expand_dims = [1] * self.ndim
            expand_dims[axis] = -1

            input_indices_axis = np.arange(self.shape[axis]).reshape(expand_dims)
            output_indices_axis = np.repeat(input_indices_axis, repeats_arr, axis=axis)

            idx = np.indices(result.grad.shape)
            idx[axis] = output_indices_axis

            np.add.at(grad_input, tuple(idx), result.grad)

        return grad_input

    return result, compute_grad


@create_ufunc_op()
def tile(self: Tensor, reps: int | Sequence[int]) -> _FuncOpReturnType:
    result = Tensor(np.tile(self.data, reps))

    def compute_grad() -> _NumPyArray:
        if self.ndim == 0:
            input_shape = (1,)
            if isinstance(reps, int):
                reps_list = (reps,)
            else:
                reps_list = tuple(reps)
                if len(reps_list) == 0:
                    reps_list = (1,)
        else:
            input_shape = np.array(self.shape)
            if isinstance(reps, int):
                reps_list = (1,) * (self.ndim - 1) + (reps,)
            else:
                reps_list = tuple(reps)
                if len(reps_list) < self.ndim:
                    reps_list = (1,) * (self.ndim - len(reps_list)) + reps_list

        reps_array = np.array(reps_list)

        reshape_dims = []
        for dim_size, rep in zip(input_shape, reps_array):
            reshape_dims.extend([rep, dim_size])

        grad_output = result.grad
        if grad_output.size != np.prod(reshape_dims):
            raise ValueError(
                f"Cannot reshape array of size {grad_output.size} "
                + f"into shape {reshape_dims}"
            )

        grad_output_reshape = grad_output.reshape(reshape_dims)
        axes_to_sum = tuple(range(0, grad_output_reshape.ndim, 2))

        return grad_output_reshape.sum(axis=axes_to_sum)

    return result, compute_grad


@create_ufunc_op()
def flatten(self: Tensor) -> _FuncOpReturnType:
    original_shape = self.shape
    result = Tensor(self.data.reshape(-1))

    def compute_grad() -> _NumPyArray:
        return result.grad.reshape(*original_shape)

    return result, compute_grad



from typing import Any
import numpy as np

from lucid._tensor import Tensor
from lucid.types import _ShapeLike, _ArrayLike


def zeros(
    shape: _ShapeLike,
    dtype: Any = np.float32,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return Tensor(np.zeros(shape, dtype=dtype), requires_grad, keep_grad)


def zeros_like(
    a: Tensor | _ArrayLike,
    dtype: Any = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    if dtype is None and hasattr(a, "dtype"):
        dtype = a.dtype
    if isinstance(a, Tensor):
        a = a.data
    return Tensor(np.zeros_like(a, dtype=dtype), requires_grad, keep_grad)


def ones(
    shape: _ShapeLike,
    dtype: Any = np.float32,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return Tensor(np.ones(shape, dtype=dtype), requires_grad, keep_grad)


def ones_like(
    a: Tensor | _ArrayLike,
    dtype: Any = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    if dtype is None and hasattr(a, "dtype"):
        dtype = a.dtype
    if isinstance(a, Tensor):
        a = a.data
    return Tensor(np.ones_like(a, dtype=dtype), requires_grad, keep_grad)


def eye(
    N: int,
    M: int | None = None,
    k: int = 0,
    dtype: Any = np.float32,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return Tensor(np.eye(N, M, k, dtype=dtype), requires_grad, keep_grad)


def diag(
    v: Tensor | _ArrayLike,
    k: int = 0,
    dtype: Any = np.float32,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    if not isinstance(v, Tensor):
        v = Tensor(v, requires_grad, keep_grad, dtype)
    return Tensor(np.diag(v.data, k), v.requires_grad, v.keep_grad)



import lucid
from lucid._tensor import Tensor

from lucid.types import _ShapeLike


def batch_norm(
    input_: Tensor,
    running_mean: Tensor,
    running_var: Tensor,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    training: bool = True,
    momentum: float = 0.1,
    eps: float = 1e-5,
) -> Tensor:
    C = input_.shape[1]
    spatial_dim = input_.ndim - 2

    if training:
        batch_mean = input_.mean(axis=(0, *range(2, input_.ndim)), keepdims=True)
        batch_var = input_.var(axis=(0, *range(2, input_.ndim)), keepdims=True)

        running_mean = momentum * batch_mean.flatten() + (1 - momentum) * running_mean
        running_var = momentum * batch_var.flatten() + (1 - momentum) * running_var

        mean = batch_mean
        var = batch_var
    else:
        mean = running_mean.reshape(1, C, *(1,) * spatial_dim)
        var = running_var.reshape(1, C, *(1,) * spatial_dim)

    normalized = (input_ - mean) / lucid.sqrt(var + eps)
    if weight is not None:
        weight = weight.reshape((1, C) + (1,) * spatial_dim)
        normalized *= weight

    if bias is not None:
        bias = bias.reshape((1, C) + (1,) * spatial_dim)
        normalized += bias

    return normalized


def layer_norm(
    input_: Tensor,
    normalized_shape: _ShapeLike,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    eps: float = 1e-5,
) -> Tensor:
    if input_.shape[-len(normalized_shape) :] != normalized_shape:
        raise ValueError(
            "Input tensor's normalized shape must match "
            + "the provided `normalized_shape`."
        )

    mean = input_.mean(axis=tuple(range(-len(normalized_shape), 0)), keepdims=True)
    var = input_.var(axis=tuple(range(-len(normalized_shape), 0)), keepdims=True)

    normalized = (input_ - mean) / lucid.sqrt(var + eps)
    if weight is not None:
        normalized *= weight.reshape(
            (1,) * (input_.ndim - len(normalized_shape)) + normalized_shape
        )
    if bias is not None:
        normalized += bias.reshape(
            (1,) * (input_.ndim - len(normalized_shape)) + normalized_shape
        )

    return normalized


def instance_norm(
    input_: Tensor,
    running_mean: Tensor,
    running_var: Tensor,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    training: bool = True,
    momentum: float = 0.1,
    eps: float = 1e-5,
) -> Tensor:
    C = input_.shape[1]
    spatial_dims = input_.shape[2:]

    if training:
        batch_mean = input_.mean(axis=(2, *range(2, input_.ndim)), keepdims=True)
        batch_var = input_.var(axis=(2, *range(2, input_.ndim)), keepdims=True)

        running_mean = (
            momentum * batch_mean.mean(axis=0).flatten() + (1 - momentum) * running_mean
        )
        running_var = (
            momentum * batch_var.mean(axis=0).flatten() + (1 - momentum) * running_var
        )

        mean = batch_mean
        var = batch_var
    else:
        mean = running_mean.reshape(1, C, *(1,) * len(spatial_dims))
        var = running_var.reshape(1, C, *(1,) * len(spatial_dims))

    normalized = (input_ - mean) / lucid.sqrt(var + eps)
    if weight is not None:
        normalized *= weight.reshape(1, C, *(1,) * len(spatial_dims))
    if bias is not None:
        normalized += bias.reshape(1, C, *(1,) * len(spatial_dims))

    return normalized



from typing import Any
import numpy as np

from lucid._func import bfunc, gfunc, ufunc
from lucid._tensor import Tensor
from lucid.types import _Scalar, _ShapeLike, _ArrayLike


def add(a: Tensor, b: Tensor) -> Tensor:
    return bfunc._add(a, b)


def sub(a: Tensor, b: Tensor) -> Tensor:
    return bfunc._sub(a, b)


def mul(a: Tensor, b: Tensor) -> Tensor:
    return bfunc._mul(a, b)


def div(a: Tensor, b: Tensor) -> Tensor:
    return bfunc._truediv(a, b)


def minimum(a: Tensor, b: Tensor) -> Tensor:
    return bfunc.minimum(a, b)


def maximum(a: Tensor, b: Tensor) -> Tensor:
    return bfunc.maximum(a, b)


def power(a: Tensor, b: Tensor) -> Tensor:
    return bfunc.power(a, b)


def dot(a: Tensor, b: Tensor) -> Tensor:
    return bfunc.dot(a, b)


def inner(a: Tensor, b: Tensor) -> Tensor:
    return bfunc.inner(a, b)


def outer(a: Tensor, b: Tensor) -> Tensor:
    return bfunc.outer(a.ravel(), b.ravel())


def matmul(a: Tensor, b: Tensor) -> Tensor:
    return bfunc.matmul(a, b)


def exp(a: Tensor) -> Tensor:
    return ufunc.exp(a)


def log(a: Tensor) -> Tensor:
    return ufunc.log(a)


def sqrt(a: Tensor) -> Tensor:
    return ufunc.sqrt(a)


def sin(a: Tensor) -> Tensor:
    return ufunc.sin(a)


def cos(a: Tensor) -> Tensor:
    return ufunc.cos(a)


def tan(a: Tensor) -> Tensor:
    return ufunc.tan(a)


def arcsin(a: Tensor) -> Tensor:
    return ufunc.arcsin(a)


def arccos(a: Tensor) -> Tensor:
    return ufunc.arccos(a)


def arctan(a: Tensor) -> Tensor:
    return ufunc.arctan(a)


def sinh(a: Tensor) -> Tensor:
    return ufunc.sinh(a)


def cosh(a: Tensor) -> Tensor:
    return ufunc.cosh(a)


def tanh(a: Tensor) -> Tensor:
    return ufunc.tanh(a)


def clip(a: Tensor, min_value: _Scalar, max_value: _Scalar) -> Tensor:
    return ufunc.clip(a, min_value, max_value)


def abs(a: Tensor) -> Tensor:
    return ufunc.abs(a)


def sign(a: Tensor) -> Tensor:
    return ufunc.sign(a)


def reciprocal(a: Tensor) -> Tensor:
    return ufunc.reciprocal(a)


def square(a: Tensor) -> Tensor:
    return ufunc.square(a)


def cube(a: Tensor) -> Tensor:
    return ufunc.cube(a)


def transpose(a: Tensor, axes: list[int] | None = None) -> Tensor:
    return ufunc.transpose(a, axes)


def sum(
    a: Tensor, axis: int | tuple[int] | None = None, keepdims: bool = False
) -> Tensor:
    return ufunc.sum(a, axis, keepdims)


def trace(a: Tensor) -> Tensor:
    return ufunc.trace(a)


def mean(
    a: Tensor, axis: int | tuple[int] | None = None, keepdims: bool = False
) -> Tensor:
    return ufunc.mean(a, axis, keepdims)


def var(
    a: Tensor, axis: int | tuple[int] | None = None, keepdims: bool = False
) -> Tensor:
    return ufunc.var(a, axis, keepdims)


def min(
    a: Tensor, axis: int | tuple[int] | None = None, keepdims: bool = False
) -> Tensor:
    return ufunc._min_or_max(a, "min", axis, keepdims)


def max(
    a: Tensor, axis: int | tuple[int] | None = None, keepdims: bool = False
) -> Tensor:
    return ufunc._min_or_max(a, "max", axis, keepdims)


def zeros(
    shape: _ShapeLike,
    dtype: Any = np.float32,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return gfunc.zeros(shape, dtype, requires_grad, keep_grad)


def zeros_like(
    a: Tensor | _ArrayLike,
    dtype: Any = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return gfunc.zeros_like(a, dtype, requires_grad, keep_grad)


def ones(
    shape: _ShapeLike,
    dtype: Any = np.float32,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return gfunc.ones(shape, dtype, requires_grad, keep_grad)


def ones_like(
    a: Tensor | _ArrayLike,
    dtype: Any = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return gfunc.ones_like(a, dtype, requires_grad, keep_grad)


def eye(
    N: int,
    M: int | None = None,
    k: int = 0,
    dtype: Any = np.float32,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return gfunc.eye(N, M, k, dtype, requires_grad, keep_grad)


def diag(
    v: Tensor | _ArrayLike,
    k: int = 0,
    dtype: Any = np.float32,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return gfunc.diag(v, k, dtype, requires_grad, keep_grad)


Tensor.__add__ = bfunc._add
Tensor.__radd__ = bfunc._radd
Tensor.__sub__ = bfunc._sub
Tensor.__rsub__ = bfunc._rsub
Tensor.__mul__ = bfunc._mul
Tensor.__rmul__ = bfunc._rmul
Tensor.__truediv__ = bfunc._truediv
Tensor.__rtruediv__ = bfunc._rtruediv
Tensor.__matmul__ = bfunc._matmul

Tensor.__eq__ = bfunc._equal
Tensor.__ne__ = bfunc._not_equal
Tensor.__gt__ = bfunc._greater
Tensor.__ge__ = bfunc._greater_or_equal
Tensor.__lt__ = bfunc._less
Tensor.__le__ = bfunc._less_or_equal

Tensor.__pow__ = ufunc._pow
Tensor.__neg__ = ufunc._neg

Tensor.T = ufunc._T
Tensor.mT = ufunc._mT
Tensor.transpose = ufunc.transpose
Tensor.dot = bfunc.dot
Tensor.matmul = bfunc._matmul
Tensor.sum = ufunc.sum
Tensor.mean = ufunc.mean
Tensor.var = ufunc.var



import functools
from typing import Callable, Tuple

import numpy as np

import lucid
from lucid._tensor import Tensor
from lucid.types import _NumPyArray

_GradFuncType = Callable[[None], _NumPyArray | Tuple[_NumPyArray, ...]]

_ReturnGradFuncPair = Tuple[Tensor, _GradFuncType]

_FuncOpReturnType = _ReturnGradFuncPair | Tuple[_ReturnGradFuncPair, ...]


def create_func_op(n_in: int | None, n_ret: int, has_gradient: bool = True) -> callable:

    def decorator(func: Callable[..., _FuncOpReturnType]) -> callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Tuple[Tensor, ...]:
            tensors: Tuple[Tensor, ...] = tuple()
            requires_grad = False

            if n_in is None:
                tensor_args = args
            else:
                if len(args) < n_in:
                    raise ValueError(
                        f"Expected at least {n_in} tensor arguments, got {len(args)}"
                    )
                tensor_args = args[:n_in]

            for arg in tensor_args:
                tensor = lucid._check_is_tensor(arg)
                tensors += (tensor,)
                requires_grad = requires_grad or tensor.requires_grad

            non_tensor_args = args[n_in:] if n_in is not None else ()
            new_args = (*tensors, *non_tensor_args)

            func_return_pairs = func(*new_args, **kwargs)

            if n_ret == 1:
                func_return_pairs = (func_return_pairs,)

            results: Tuple[Tensor, ...] = tuple()
            for result, compute_grad in func_return_pairs:
                result.requires_grad = requires_grad and has_gradient
                results += (result,)

                def _backward_op(*, _func: Callable = compute_grad) -> None:
                    grads = _func()
                    if n_in == 1 or not isinstance(grads, tuple):
                        grads = (grads,)

                    if len(grads) != len(tensors):
                        raise ValueError(
                            f"Expected {len(tensors)} gradients, got {len(grads)}."
                        )

                    for tensor, grad in zip(tensors, grads):
                        new_grad = lucid._match_grad_shape(tensor.data, grad)
                        lucid._set_tensor_grad(tensor, new_grad)

                if not lucid.grad_enabled():
                    continue

                if result.requires_grad:
                    result._backward_op = _backward_op
                    result._prev = tensors

            return results if n_ret > 1 else results[0]

        return wrapper

    return decorator


def create_bfunc_op(has_gradient: bool = True) -> callable:
    return create_func_op(n_in=2, n_ret=1, has_gradient=has_gradient)


def create_ufunc_op(has_gradient: bool = True) -> callable:
    return create_func_op(n_in=1, n_ret=1, has_gradient=has_gradient)


def create_mfunc_op(has_gradient: bool = True) -> callable:
    return create_func_op(n_in=None, n_ret=1, has_gradient=has_gradient)



import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from typing import Any, Literal


__all__ = ["Conv1d", "Conv2d", "Conv3d"]


def _single_to_tuple(value: Any, times: int) -> tuple[Any, ...]:
    if isinstance(value, tuple):
        return value
    return (value,) * times


_PaddingStr = Literal["same", "valid"]


class _ConvNd(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...],
        padding: _PaddingStr | int | tuple[int, ...],
        dilation: int | tuple[int, ...],
        groups: int,
        bias: bool,
        *,
        D: int,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.groups = groups

        self.kernel_size = _single_to_tuple(kernel_size, D)
        self.stride = _single_to_tuple(stride, D)
        self.dilation = _single_to_tuple(dilation, D)

        if isinstance(padding, str):
            if padding == "same":
                self.padding = tuple(
                    (self.dilation[i] * (self.kernel_size[i] - 1)) // 2
                    for i in range(D)
                )
            elif padding == "valid":
                self.padding = (0,) * D
            else:
                raise ValueError(f"Unknown padding string: {padding}")
        else:
            self.padding = _single_to_tuple(padding, D)

        if groups <= 0:
            raise ValueError("groups must be a positive integer.")
        if in_channels % groups != 0:
            raise ValueError("in_channels must be divisible by groups.")
        if out_channels % groups != 0:
            raise ValueError("out_channels mube be divisible by groups.")

        weight_ = lucid.random.randn(
            out_channels, in_channels // groups, *self.kernel_size
        )
        self.weight = nn.Parameter(weight_ * 0.01)

        if bias:
            bias_ = lucid.zeros((out_channels,))
            self.bias = nn.Parameter(bias_)
        else:
            self.bias = None


class Conv1d(_ConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        bias: bool = True,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            D=1,
        )

    def _conv_forward(
        self, input_: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:
        return F.conv1d(
            input_, weight, bias, self.stride, self.padding, self.dilation, self.groups
        )

    def forward(self, input_: Tensor) -> Tensor:
        return self._conv_forward(input_, self.weight, self.bias)


class Conv2d(_ConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        bias: bool = True,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            D=2,
        )

    def _conv_forward(
        self, input_: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:
        return F.conv2d(
            input_, weight, bias, self.stride, self.padding, self.dilation, self.groups
        )

    def forward(self, input_: Tensor) -> Tensor:
        return self._conv_forward(input_, self.weight, self.bias)


class Conv3d(_ConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        bias: bool = True,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            D=3,
        )

    def _conv_forward(
        self, input_: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:
        return F.conv3d(
            input_, weight, bias, self.stride, self.padding, self.dilation, self.groups
        )

    def forward(self, input_: Tensor) -> Tensor:
        return self._conv_forward(input_, self.weight, self.bias)



from typing import Sequence
from lucid._tensor import Tensor
from lucid.types import _ShapeLike, _ArrayLikeInt

from lucid._util import func


def reshape(a: Tensor, shape: _ShapeLike) -> Tensor:
    return func._reshape(a, shape)


def squeeze(a: Tensor, axis: _ShapeLike | None = None) -> Tensor:
    return func.squeeze(a, axis)


def unsqueeze(a: Tensor, axis: _ShapeLike) -> Tensor:
    return func.unsqueeze(a, axis)


def ravel(a: Tensor) -> Tensor:
    return func.ravel(a)


def stack(arr: tuple[Tensor, ...], axis: int = 0) -> Tensor:
    return func.stack(*arr, axis=axis)


def hstack(arr: tuple[Tensor, ...]) -> Tensor:
    return func.hstack(*arr)


def vstack(arr: tuple[Tensor, ...]) -> Tensor:
    return func.vstack(*arr)


def concatenate(arr: tuple[Tensor, ...], axis: int = 0) -> Tensor:
    return func.concatenate(*arr, axis=axis)


def pad(a: Tensor, pad_width: _ArrayLikeInt) -> Tensor:
    return func.pad(a, pad_width)


def repeat(a: Tensor, repeats: int | Sequence[int], axis: int | None = None) -> Tensor:
    return func.repeat(a, repeats, axis=axis)


def tile(a: Tensor, reps: int | Sequence[int]) -> Tensor:
    return func.tile(a, reps)


def flatten(a: Tensor) -> Tensor:
    return func.flatten(a)


Tensor.reshape = func._reshape_inplace
Tensor.squeeze = func.squeeze
Tensor.unsqueeze = func.unsqueeze
Tensor.ravel = func.ravel
Tensor.pad = func.pad
Tensor.repeat = func.repeat
Tensor.tile = func.tile
Tensor.flatten = func.flatten



import numpy as np

from lucid._tensor import Tensor
from lucid.types import _ShapeLike


def seed(seed: int) -> None:
    np.random.seed(seed)


def rand(*shape: int, requires_grad: bool = False, keep_grad: bool = False) -> Tensor:
    return Tensor(np.random.rand(*shape), requires_grad, keep_grad)


def randint(
    low: int,
    high: int | None,
    size: int | _ShapeLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return Tensor(np.random.randint(low, high, size), requires_grad, keep_grad)


def randn(*shape: int, requires_grad: bool = False, keep_grad: bool = False) -> Tensor:
    return Tensor(np.random.randn(*shape), requires_grad, keep_grad)



from typing import Any, Iterator, Self
from collections import OrderedDict
import numpy as np

from lucid._tensor import Tensor
from lucid.types import _ArrayOrScalar


class Parameter(Tensor):
    def __init__(self, data: Tensor | _ArrayOrScalar, dtype=np.float32):
        if isinstance(data, Tensor):
            data = data.data
        super().__init__(data, requires_grad=True, keep_grad=True, dtype=dtype)


class Module:
    def __init__(self) -> None:
        object.__setattr__(self, "_parameters", OrderedDict())
        object.__setattr__(self, "_modules", OrderedDict())

    def __setattr__(self, name: str, value: Any) -> None:
        if isinstance(value, Parameter):
            if name in self._modules:
                del self._modules[name]
            self._parameters[name] = value

        elif isinstance(value, Module):
            if name in self._parameters:
                del self._parameters[name]
            self._modules[name] = value

        else:
            if name in self._parameters:
                del self._parameters[name]
            if name in self._modules:
                del self._modules[name]

        super().__setattr__(name, value)

    def add_module(self, name: str, module: Self) -> None:
        if not isinstance(module, Module) and module is not None:
            raise TypeError(f"{module} is not a Module.")

        self.__setattr__(name, module)

    def register_parameter(self, name: str, param: Parameter) -> None:
        if not isinstance(param, Parameter) and param is not None:
            raise TypeError(f"{param} is not a Parameter.")

        self.__setattr__(name, param)

    def forward(self) -> Tensor | tuple[Tensor, ...]:
        raise NotImplementedError(
            "The forward method must be implemented by the subclass."
        )

    def parameters(self, recurse: bool = True) -> Iterator:
        for _, param in self._parameters.items():
            yield param
        if recurse:
            for module in self._modules.values():
                yield from module.parameters(recurse=recurse)

    def modules(self) -> Iterator:
        yield self
        for module in self._modules.values():
            yield from module.modules()

    def state_dict(
        self,
        destination: OrderedDict | None = None,
        prefix: str = "",
        keep_vars: bool = False,
    ) -> dict[str, Parameter]:
        if destination is None:
            destination = OrderedDict()

        for name, param in self._parameters.items():
            destination[prefix + name] = param

        for name, module in self._modules.items():
            module.state_dict(
                destination=destination, prefix=prefix + name + ".", keep_vars=keep_vars
            )
        return destination

    def load_state_dict(
        self, state_dict: dict[str, Parameter], strict: bool = True
    ) -> None:
        own_state = self.state_dict()

        if strict and set(own_state.keys()) != set(state_dict.keys()):
            missing = set(own_state.keys()) - set(state_dict.keys())
            unexpected = set(state_dict.keys()) - set(own_state.keys())
            msg = ""

            if missing:
                msg += f"Missing keys in state_dict: {missing}\n"
            if unexpected:
                msg += f"Unexpected keys in state_dict: {unexpected}\n"
            if msg:
                raise KeyError("Error(s) in loading state_dict:\n" + msg)

        for key, param in state_dict.items():
            splits = key.split(".")
            obj = self
            for split in splits[:-1]:
                obj = obj._modules[split]

            final_name = splits[-1]
            setattr(obj, final_name, param)

    def __call__(self, *args: Any, **kwargs: Any) -> Tensor | tuple[Tensor, ...]:
        return self.forward(*args, **kwargs)



from typing import Sequence
import numpy as np


_Scalar = int | float
_NumPyArray = np.ndarray
_ArrayOrScalar = _Scalar | list[_Scalar] | _NumPyArray

_ShapeLike = list[int] | tuple[int]

_ArrayLike = list | _NumPyArray
_ArrayLikeInt = int | Sequence[int | tuple[int, int]]



import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor


__all__ = ["Identity", "Linear", "Bilinear"]


class Identity(nn.Module):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return input_


class Linear(nn.Module):
    def __init__(self, in_features: int, out_features: int, bias: bool = True) -> None:
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features

        weight_ = lucid.random.randn(out_features, in_features)
        self.weight = nn.Parameter(weight_)

        if bias:
            bias_ = lucid.zeros((1, out_features))
            self.bias = nn.Parameter(bias_)
        else:
            self.bias = None

    def forward(self, input_: Tensor) -> Tensor:
        return F.linear(input_, self.weight, self.bias)


class Bilinear(nn.Module):
    def __init__(
        self, in1_features: int, in2_features: int, out_features: int, bias: bool = True
    ) -> None:
        super().__init__()
        self.in1_features = in1_features
        self.in2_features = in2_features
        self.out_features = out_features

        weight_ = lucid.random.randn(out_features, in1_features, in2_features)
        self.weight = nn.Parameter(weight_)

        if bias:
            bias_ = lucid.zeros((1, out_features))
            self.bias = nn.Parameter(bias_)
        else:
            self.bias = None

    def forward(self, input_1: Tensor, input_2: Tensor) -> Tensor:
        return F.bilinear(input_1, input_2, self.weight, self.bias)



from lucid.random import _func

from lucid._tensor import Tensor
from lucid.types import _ShapeLike


def seed(seed: int) -> None:
    return _func.seed(seed)


def rand(*shape: int, requires_grad: bool = False, keep_grad: bool = False) -> Tensor:
    return _func.rand(*shape, requires_grad=requires_grad, keep_grad=keep_grad)


def randint(
    low: int,
    high: int | None,
    size: int | _ShapeLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
) -> Tensor:
    return _func.randint(low, high, size, requires_grad, keep_grad)


def randn(*shape: int, requires_grad: bool = False, keep_grad: bool = False) -> Tensor:
    return _func.randn(*shape, requires_grad=requires_grad, keep_grad=keep_grad)



import math
import lucid
from lucid._tensor import Tensor


def _pad_input(input_: Tensor, padding: tuple[int, ...]) -> Tensor:
    pad_config = [(0, 0), (0, 0)] + [(pad, pad) for pad in padding]
    return lucid.pad(input_, pad_config)


def _pool1d(
    input_: Tensor,
    kernel_size: int | tuple[int],
    stride: int | tuple[int],
    padding: int | tuple[int],
) -> Tensor:
    if isinstance(kernel_size, int):
        kernel_size = (kernel_size,)
    if isinstance(stride, int):
        stride = (stride,)
    if isinstance(padding, int):
        padding = (padding,)

    _, _, L = input_.shape
    (kernel,) = kernel_size
    (stride_,) = stride
    (pad_,) = padding

    out_L = math.floor((L + 2 * pad_ - kernel) / stride_ + 1)
    padded_input = _pad_input(input_, padding)

    patches = []
    for i in range(kernel):
        start = i
        end = start + stride_ * out_L

        patch = padded_input[..., start:end:stride_]
        patches.append(patch)

    return lucid.stack(patches, axis=-1)


def _pool2d(
    input_: Tensor,
    kernel_size: int | tuple[int, int],
    stride: int | tuple[int, int],
    padding: int | tuple[int, int],
) -> Tensor:
    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)
    if isinstance(stride, int):
        stride = (stride, stride)
    if isinstance(padding, int):
        padding = (padding, padding)

    _, _, H, W = input_.shape
    kernel_h, kernel_w = kernel_size
    stride_h, stride_w = stride
    pad_h, pad_w = padding

    out_H = math.floor((H + 2 * pad_h - kernel_h) / stride_h + 1)
    out_W = math.floor((W + 2 * pad_w - kernel_w) / stride_w + 1)
    padded_input = _pad_input(input_, padding)

    patches = []
    for i in range(kernel_h):
        for j in range(kernel_w):
            start_h = i
            end_h = start_h + stride_h * out_H
            start_w = j
            end_w = start_w + stride_w * out_W

            patch = padded_input[
                ...,
                start_h:end_h:stride_h,
                start_w:end_w:stride_w,
            ]
            patches.append(patch)

    return lucid.stack(patches, axis=-1)


def _pool3d(
    input_: Tensor,
    kernel_size: int | tuple[int, int, int],
    stride: int | tuple[int, int, int],
    padding: int | tuple[int, int, int],
) -> Tensor:
    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size, kernel_size)
    if isinstance(stride, int):
        stride = (stride, stride, stride)
    if isinstance(padding, int):
        padding = (padding, padding, padding)

    _, _, D, H, W = input_.shape
    kernel_d, kernel_h, kernel_w = kernel_size
    stride_d, stride_h, stride_w = stride
    pad_d, pad_h, pad_w = padding

    out_D = math.floor((D + 2 * pad_d - kernel_d) / stride_d + 1)
    out_H = math.floor((H + 2 * pad_h - kernel_h) / stride_h + 1)
    out_W = math.floor((W + 2 * pad_w - kernel_w) / stride_w + 1)
    padded_input = _pad_input(input_, padding)

    patches = []
    for i in range(kernel_d):
        for j in range(kernel_h):
            for k in range(kernel_w):
                start_d = i
                end_d = start_d + stride_d * out_D
                start_h = j
                end_h = start_h + stride_h * out_H
                start_w = k
                end_w = start_w + stride_w * out_W

                patch = padded_input[
                    ...,
                    start_d:end_d:stride_d,
                    start_h:end_h:stride_h,
                    start_w:end_w:stride_w,
                ]
                patches.append(patch)

    return lucid.stack(patches, axis=-1)


def avg_pool1d(
    input_: Tensor,
    kernel_size: int | tuple[int] = 1,
    stride: int | tuple[int] = 1,
    padding: int | tuple[int] = 0,
) -> Tensor:
    return (
        _pool1d(input_, kernel_size, stride, padding)
        .mean(axis=-1, keepdims=True)
        .squeeze()
    )


def avg_pool2d(
    input_: Tensor,
    kernel_size: int | tuple[int, int] = 1,
    stride: int | tuple[int, int] = 1,
    padding: int | tuple[int, int] = 0,
) -> Tensor:
    return (
        _pool2d(input_, kernel_size, stride, padding)
        .mean(axis=-1, keepdims=True)
        .squeeze()
    )


def avg_pool3d(
    input_: Tensor,
    kernel_size: int | tuple[int, int, int] = 1,
    stride: int | tuple[int, int, int] = 1,
    padding: int | tuple[int, int, int] = 0,
) -> Tensor:
    return (
        _pool3d(input_, kernel_size, stride, padding)
        .mean(axis=-1, keepdims=True)
        .squeeze()
    )


def max_pool1d(
    input_: Tensor,
    kernel_size: int | tuple[int] = 1,
    stride: int | tuple[int] = 1,
    padding: int | tuple[int] = 0,
) -> Tensor:
    return lucid.max(_pool1d(input_, kernel_size, stride, padding), axis=-1)


def max_pool2d(
    input_: Tensor,
    kernel_size: int | tuple[int, int] = 1,
    stride: int | tuple[int, int] = 1,
    padding: int | tuple[int, int] = 0,
) -> Tensor:
    return lucid.max(_pool2d(input_, kernel_size, stride, padding), axis=-1)


def max_pool3d(
    input_: Tensor,
    kernel_size: int | tuple[int, int, int] = 1,
    stride: int | tuple[int, int, int] = 1,
    padding: int | tuple[int, int, int] = 0,
) -> Tensor:
    return lucid.max(_pool3d(input_, kernel_size, stride, padding), axis=-1)



"""
# `Lucid `

**Lucid** is an educational deep learning framework developed to help users understand 
the underlying mechanics of deep learning models and tensor operations. 

It is designed to provide a simple yet powerful environment to experiment with neural networks, 
optimization, and backpropagation using only `NumPy`. 

Lucid is ideal for those who want to learn about the inner workings of deep learning 
algorithms and operations without the complexity of high-level frameworks.

[📑 Lucid Documentation](https://chanlumerico.github.io/lucid/build/html/index.html)
"""

from contextlib import contextmanager
from typing import Any, Generator, SupportsIndex
import numpy as np

from lucid._tensor import Tensor
from lucid._func import *
from lucid._util import *

from lucid.types import _ArrayOrScalar, _NumPyArray, _ArrayLike, _ShapeLike

import lucid.linalg as linalg
import lucid.random as random
import lucid.nn as nn

_grad_enabled: bool = True

newaxis = np.newaxis

pi = np.pi
inf = np.inf


def tensor(
    data: Tensor | _ArrayOrScalar,
    requires_grad: bool = False,
    keep_grad: bool = False,
    dtype: Any = np.float32,
) -> Tensor:
    if isinstance(data, Tensor):
        data = data.data
    return Tensor(data, requires_grad, keep_grad, dtype)


def to_tensor(
    a: _ArrayLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
    dtype: Any = np.float32,
) -> Tensor:
    return tensor(a, requires_grad, keep_grad, dtype)


@contextmanager
def no_grad() -> Generator:
    global _grad_enabled
    prev_state = _grad_enabled
    _grad_enabled = False
    try:
        yield
    finally:
        _grad_enabled = prev_state


def grad_enabled() -> bool:
    return _grad_enabled


def shape(a: Tensor | _NumPyArray) -> _ShapeLike:
    if hasattr(a, "shape"):
        return a.shape

    raise ValueError(f"The argument must be a Tensor or a NumPy array.")


def _set_tensor_grad(
    tensor: Tensor, grad: _NumPyArray, at: SupportsIndex = ...
) -> None:
    if tensor.requires_grad:
        if tensor.grad is None:
            tensor.grad = grad
        else:
            tensor.grad[at] = tensor.grad[at] + grad


def _check_is_tensor(any: Tensor | _ArrayOrScalar) -> Tensor:
    if not isinstance(any, Tensor):
        return Tensor(any)
    return any


def _match_grad_shape(data: _NumPyArray, grad: _NumPyArray) -> _NumPyArray:
    if data.shape == grad.shape:
        return grad

    if data.ndim == 0:
        return grad.sum()

    matched_grad = grad
    if data.ndim > grad.ndim:
        matched_grad = grad.reshape(data.shape)

    elif data.ndim < grad.ndim:
        squeezed = grad.squeeze()
        matched_grad = squeezed.reshape(data.shape)

    else:
        if data.size == grad.size:
            matched_grad = grad
        elif data.size < grad.size:
            axis = []
            if data.ndim == 0:
                axis.extend(range(grad.ndim))
            else:
                for ax in range(data.ndim):
                    if data.shape[ax] != grad.shape[ax] and data.shape[ax] == 1:
                        axis.append(ax)

            matched_grad = np.sum(grad, axis=tuple(axis)).reshape(data.shape)
        else:
            if data.size % grad.size == 0:
                matched_grad = np.broadcast_to(grad, data.shape)

    return matched_grad



from typing import Any, Iterator, Optional, Self, SupportsIndex
import numpy as np
import lucid
from lucid._tensor.tensor_ops import _TensorOps
from lucid.types import _ArrayOrScalar, _NumPyArray, _Scalar


class Tensor(_TensorOps):
    def __init__(
        self,
        data: _ArrayOrScalar,
        requires_grad: bool = False,
        keep_grad: bool = False,
        dtype: Any = np.float32,
    ) -> None:
        if not isinstance(data, _NumPyArray):
            self.data = np.array(data, dtype=dtype)
        else:
            self.data = data

        self.requires_grad = requires_grad and lucid.grad_enabled()
        self.keep_grad = keep_grad
        self.dtype = self.data.dtype

        self.grad: Optional[_NumPyArray] = None
        self._backward_op: callable = lambda: None
        self._prev: list[Tensor] = []

    @property
    def is_leaf(self) -> bool:
        return self.requires_grad and len(self._prev) == 0

    def backward(self, keep_grad: bool = False) -> None:
        if self.grad is None:
            self.grad = np.ones_like(self.data)

        visited = set()
        topo_order: list[Tensor] = []

        def _build_topo(tensor: Tensor) -> None:
            if tensor not in visited:
                visited.add(tensor)
                for parent in tensor._prev:
                    _build_topo(parent)
                topo_order.append(tensor)

        _build_topo(self)
        topo_order.reverse()

        for tensor in topo_order:
            tensor._backward_op()
            if not tensor.is_leaf and not keep_grad:
                self.zero_grad()

    @property
    def shape(self) -> tuple[int, ...]:
        return self.data.shape

    @property
    def ndim(self) -> int:
        return self.data.ndim

    @property
    def size(self) -> int:
        return self.data.size

    def item(self) -> _Scalar:
        if self.ndim != 0:
            raise ValueError(
                "Tensor must be 0-dimensional(scalar) to pop its item.",
            )
        item = self.data
        if item % 1 == 0:
            return int(item)
        else:
            return float(item)

    def zero_grad(self) -> None:
        if not self.keep_grad:
            self.grad = None

    def __getitem__(self, idx: SupportsIndex) -> Self:
        sliced_data = self.data[idx]
        new_tensor = Tensor(sliced_data, requires_grad=self.requires_grad)

        def _backward_op() -> None:
            if self.grad is None:
                self.grad = np.zeros_like(self.data)

            new_grad = lucid._match_grad_shape(self.data[idx], new_tensor.grad)
            lucid._set_tensor_grad(self, new_grad, at=idx)

        if self.requires_grad:
            new_tensor._backward_op = _backward_op
            new_tensor._prev = [self]

        return new_tensor

    def __setitem__(self, idx: SupportsIndex, value: Self | _ArrayOrScalar) -> None:
        if self.requires_grad:
            raise RuntimeError(
                "Cannot perform in-place item setting on a "
                + "Tensor that requires gradients. "
            )

        if not isinstance(value, Tensor):
            value = Tensor(value)
        self.data[idx] = value.data

    def __iter__(self) -> Iterator[Self]:
        for i in range(self.shape[0]):
            yield self[i]

    def __repr__(self) -> str:
        return f"Tensor({self.data}, grad={self.grad})"

    def __str__(self) -> str:
        return str(self.data)

    def __hash__(self) -> int:
        return hash(id(self))



from typing import Literal, Optional
import numpy as np

from lucid._tensor import Tensor
from lucid._backend import create_ufunc_op, _FuncOpReturnType
from lucid.types import _NumPyArray, _Scalar


@create_ufunc_op()
def _pow(self: Tensor, exp: _Scalar) -> _FuncOpReturnType:
    result = Tensor(self.data**exp)

    def compute_grad() -> _NumPyArray:
        return (exp * self.data ** (exp - 1)) * result.grad

    return result, compute_grad


@create_ufunc_op()
def _neg(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(-self.data)

    def compute_grad() -> _NumPyArray:
        return -result.grad

    return result, compute_grad


@create_ufunc_op()
def exp(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.exp(self.data))

    def compute_grad() -> _NumPyArray:
        return result.data * result.grad

    return result, compute_grad


@create_ufunc_op()
def log(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.log(self.data))

    def compute_grad() -> _NumPyArray:
        return (1 / self.data) * result.grad

    return result, compute_grad


@create_ufunc_op()
def sqrt(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.sqrt(self.data))

    def compute_grad() -> _NumPyArray:
        return (0.5 / result.data) * result.grad

    return result, compute_grad


@create_ufunc_op()
def sin(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.sin(self.data))

    def compute_grad() -> _NumPyArray:
        return np.cos(self.data) * result.grad

    return result, compute_grad


@create_ufunc_op()
def cos(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.cos(self.data))

    def compute_grad() -> _NumPyArray:
        return -np.sin(self.data) * result.grad

    return result, compute_grad


@create_ufunc_op()
def tan(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.tan(self.data))

    def compute_grad() -> _NumPyArray:
        return (1 / (np.cos(self.data) ** 2)) * result.grad

    return result, compute_grad


@create_ufunc_op()
def arcsin(self: Tensor) -> Tensor:
    result = Tensor(np.arcsin(self.data))

    def compute_grad() -> _NumPyArray:
        return (1 / np.sqrt(1 - self.data**2)) * result.grad

    return result, compute_grad


@create_ufunc_op()
def arccos(self: Tensor) -> Tensor:
    result = Tensor(np.arccos(self.data))

    def compute_grad() -> _NumPyArray:
        return (-1 / np.sqrt(1 - self.data**2)) * result.grad

    return result, compute_grad


@create_ufunc_op()
def arctan(self: Tensor) -> Tensor:
    result = Tensor(np.arctan(self.data))

    def compute_grad() -> _NumPyArray:
        return (1 / (1 + self.data**2)) * result.grad

    return result, compute_grad


@create_ufunc_op()
def sinh(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.sinh(self.data))

    def compute_grad() -> _NumPyArray:
        return np.cosh(self.data) * result.grad

    return result, compute_grad


@create_ufunc_op()
def cosh(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.cosh(self.data))

    def compute_grad() -> _NumPyArray:
        return np.sinh(self.data) * result.grad

    return result, compute_grad


@create_ufunc_op()
def tanh(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.tanh(self.data))

    def compute_grad() -> _NumPyArray:
        return (1 - np.tanh(self.data) ** 2) * result.grad

    return result, compute_grad


@create_ufunc_op()
def clip(self: Tensor, min_value: float, max_value: float) -> _FuncOpReturnType:
    result = Tensor(np.clip(self.data, min_value, max_value))

    def compute_grad() -> _NumPyArray:
        grad = np.ones_like(self.data)
        grad[self.data < min_value] = 0
        grad[self.data > max_value] = 0
        return grad * result.grad

    return result, compute_grad


@create_ufunc_op()
def abs(self: Tensor) -> Tensor:
    result = Tensor(np.abs(self.data))

    def compute_grad() -> _NumPyArray:
        return np.where(self.data >= 0, 1, -1) * result.grad

    return result, compute_grad


@create_ufunc_op(has_gradient=False)
def sign(self: Tensor) -> Tensor:
    result = Tensor(np.sign(self.data))

    def compute_grad() -> _NumPyArray:
        return np.array(0.0)

    return result, compute_grad


@create_ufunc_op()
def reciprocal(self: Tensor) -> Tensor:
    result = Tensor(1 / self.data)

    def compute_grad() -> _NumPyArray:
        return (-1 / (self.data**2)) * result.grad

    return result, compute_grad


@create_ufunc_op()
def square(self: Tensor) -> Tensor:
    result = Tensor(np.square(self.data))

    def compute_grad() -> _NumPyArray:
        return 2 * self.data * result.grad

    return result, compute_grad


@create_ufunc_op()
def cube(self: Tensor) -> Tensor:
    result = Tensor(self.data**3)

    def compute_grad() -> _NumPyArray:
        return 3 * self.data**2 * result.grad

    return result, compute_grad


@property
@create_ufunc_op()
def _T(self: Tensor) -> Tensor:
    result = Tensor(self.data.T)

    def compute_grad() -> _NumPyArray:
        return result.grad.T

    return result, compute_grad


@property
@create_ufunc_op()
def _mT(self: Tensor) -> Tensor:
    result = Tensor(self.data.mT)

    def compute_grad() -> _NumPyArray:
        return result.grad.mT

    return result, compute_grad


@create_ufunc_op()
def transpose(self: Tensor, axes: Optional[list[int]] = None) -> _FuncOpReturnType:
    if axes is None:
        axes = list(reversed(range(self.ndim)))
    result = Tensor(np.transpose(self.data, axes))

    def compute_grad() -> _NumPyArray:
        return np.transpose(result.grad, np.argsort(axes))

    return result, compute_grad


@create_ufunc_op()
def sum(
    self: Tensor, axis: int | tuple[int] | None = None, keepdims: bool = False
) -> _FuncOpReturnType:
    result = Tensor(np.sum(self.data, axis=axis, keepdims=keepdims))

    def compute_grad() -> _NumPyArray:
        if axis is None:
            grad = np.ones_like(self.data) * result.grad
        else:
            grad_shape = list(result.grad.shape)
            if not keepdims:
                axis_tuple = axis if isinstance(axis, tuple) else (axis,)
                for ax in axis_tuple:
                    grad_shape.insert(ax, 1)

            grad = np.reshape(result.grad, grad_shape)

        return grad

    return result, compute_grad


@create_ufunc_op()
def trace(self: Tensor) -> _FuncOpReturnType:
    result = Tensor(np.trace(self.data))

    def compute_grad() -> _NumPyArray:
        grad = np.zeros_like(self.data)
        np.fill_diagonal(grad, 1)
        return grad * result.grad

    return result, compute_grad


@create_ufunc_op()
def mean(
    self: Tensor, axis: int | tuple[int] | None = None, keepdims: bool = False
) -> _FuncOpReturnType:
    result = Tensor(np.mean(self.data, axis=axis, keepdims=keepdims))

    def compute_grad() -> _NumPyArray:
        if axis is None:
            count = self.data.size
            grad = np.ones_like(self.data) * result.grad
        else:
            axis_tuple = axis if isinstance(axis, tuple) else (axis,)
            count = np.prod([self.shape[ax] for ax in axis_tuple])

            grad_shape = list(result.grad.shape)
            if not keepdims:
                for ax in sorted(axis_tuple):
                    grad_shape.insert(ax, 1)

            grad = np.reshape(result.grad, grad_shape)

        return grad / count

    return result, compute_grad


@create_ufunc_op()
def var(
    self: Tensor, axis: int | tuple[int] | None = None, keepdims: bool = False
) -> _FuncOpReturnType:
    result = Tensor(np.var(self.data, axis=axis, keepdims=keepdims))

    def compute_grad() -> _NumPyArray:
        if axis is None:
            count = self.data.size
        else:
            axis_tuple = axis if isinstance(axis, tuple) else (axis,)
            count = np.prod([self.data.shape[ax] for ax in axis_tuple])

        mean_val = np.mean(self.data, axis=axis, keepdims=True)
        grad = (2 / count) * (self.data - mean_val) * result.grad

        if axis is not None and not keepdims:
            grad_shape = list(result.grad.shape)
            for ax in sorted(axis_tuple):
                grad_shape.insert(ax, 1)
            grad = np.reshape(grad, grad_shape)

        return grad

    return result, compute_grad


@create_ufunc_op()
def _min_or_max(
    self: Tensor,
    mode: Literal["min", "max"],
    axis: int | tuple[int] | None = None,
    keepdims: bool = False,
) -> Tensor:
    if mode == "max":
        data = np.max(self.data, axis=axis, keepdims=keepdims)
    else:
        data = np.min(self.data, axis=axis, keepdims=keepdims)
    result = Tensor(data)

    def compute_grad() -> _NumPyArray:
        grad = result.grad
        if not keepdims and axis is not None:
            if isinstance(axis, tuple):
                for ax in sorted(axis):
                    grad = np.expand_dims(grad, axis=ax)
            else:
                grad = np.expand_dims(grad, axis=axis)

        if keepdims:
            result_expanded = result.data
        else:
            if axis is None:
                result_expanded = result.data.reshape((1,) * self.data.ndim)
            else:
                if isinstance(axis, tuple):
                    result_expanded = result.data
                    for ax in sorted(axis):
                        result_expanded = np.expand_dims(result_expanded, axis=ax)
                else:
                    result_expanded = np.expand_dims(result.data, axis=axis)

        mask = self.data == result_expanded
        counts = np.sum(mask, axis=axis, keepdims=True)
        counts = np.where(counts == 0, 1, counts)

        return mask * grad / counts

    return result, compute_grad



import itertools
from typing import Tuple, Optional

import lucid
from lucid._tensor import Tensor


def unfold(
    input_: Tensor,
    filter_size: Tuple[int, ...],
    stride: Tuple[int, ...],
    padding: Tuple[int, ...],
    dilation: Tuple[int, ...],
) -> Tensor:
    input_shape = input_.shape
    if len(input_shape) < 2:
        raise ValueError(
            "Input tensor must have at least 2 dimensions (N and C).",
        )

    N, C, *spatial_dims = input_shape
    D = len(spatial_dims)

    if not (len(filter_size) == len(stride) == len(padding) == len(dilation) == D):
        raise ValueError(
            "filter_size, stride, padding, and dilation must have the same "
            "length as the number of spatial dimensions."
        )

    out_dims = []
    for i in range(D):
        effective_filter_size = dilation[i] * (filter_size[i] - 1) + 1
        out_dim = (
            (spatial_dims[i] + 2 * padding[i] - effective_filter_size) // stride[i]
        ) + 1
        if out_dim <= 0:
            raise ValueError(
                f"Calculated output dimension is non-positive "
                f"for spatial dimension {i}: {out_dim}."
            )
        out_dims.append(out_dim)

    pad_config = [(0, 0), (0, 0)] + [(padding[i], padding[i]) for i in range(D)]
    padded_input = lucid.pad(input_, pad_config)

    filter_offsets = list(itertools.product(*[range(fs) for fs in filter_size]))

    patches = []
    for offset in filter_offsets:
        slices = [slice(None), slice(None)]

        for d in range(D):
            start = offset[d] * dilation[d]
            end = start + stride[d] * out_dims[d]
            step = stride[d]
            slices.append(slice(start, end, step))

        patch = padded_input[tuple(slices)]
        patch = patch.unsqueeze(axis=2 + D)
        patches.append(patch)

    col = lucid.concatenate(patches, axis=2 + D)

    new_shape = [N, C] + list(filter_size) + list(out_dims)
    col = col.reshape(new_shape)

    permute_order = [0] + list(range(2 + D, 2 + D + D)) + [1] + list(range(2, 2 + D))
    col = col.transpose(permute_order)

    N_times_out = N
    for od in out_dims:
        N_times_out *= od

    C_times_filter = C
    for fs in filter_size:
        C_times_filter *= fs

    col = col.reshape((N_times_out, C_times_filter))
    return col


def _conv(
    input_: Tensor,
    weight: Tensor,
    bias: Optional[Tensor],
    stride: Tuple[int, ...],
    padding: Tuple[int, ...],
    dilation: Tuple[int, ...],
    groups: int = 1,
) -> Tensor:
    N, C_in, *input_spatial = input_.shape
    C_out, C_in_div_g, *filter_size = weight.shape
    D = len(filter_size)

    if C_in % groups != 0:
        raise ValueError("Number of input channels is not divisible by groups.")
    if C_out % groups != 0:
        raise ValueError("Number of output channels is not divisible by groups.")
    if (C_in_div_g * groups) != C_in:
        raise ValueError("Weight shape is inconsistent with given number of groups.")

    out_dims = []
    for i in range(D):
        effective_filter_size = dilation[i] * (filter_size[i] - 1) + 1
        out_dim = (
            (input_spatial[i] + 2 * padding[i] - effective_filter_size) // stride[i]
        ) + 1
        if out_dim <= 0:
            raise ValueError(
                f"Calculated output dimension is non-positive "
                + f"for dimension {i}: {out_dim}."
            )
        out_dims.append(out_dim)

    col = unfold(input_, filter_size, stride, padding, dilation)

    prod_filter_size = 1
    for fs in filter_size:
        prod_filter_size *= fs

    C_in_group = C_in // groups
    C_out_group = C_out // groups

    weight_reshape = weight.reshape(groups, C_out_group, C_in_group * prod_filter_size)

    N_out_product = N
    for od in out_dims:
        N_out_product *= od

    col_reshape = col.reshape(N_out_product, groups, C_in_group * prod_filter_size)

    out_groups = []
    for g in range(groups):
        col_g = col_reshape[:, g, :]
        w_g = weight_reshape[g]

        out_g = col_g @ w_g.T
        out_groups.append(out_g)

    out_concated = lucid.concatenate(out_groups, axis=1)
    out = out_concated.reshape([N, C_out] + out_dims)

    if bias is not None:
        bias_shape = [1, C_out] + [1] * D
        out += bias.reshape(tuple(bias_shape))

    return out


def conv1d(
    input_: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: int | Tuple[int, ...] = 1,
    padding: int | Tuple[int, ...] = 0,
    dilation: int | Tuple[int, ...] = 1,
    groups: int = 1,
) -> Tensor:
    if isinstance(stride, int):
        stride = (stride,)
    if isinstance(padding, int):
        padding = (padding,)
    if isinstance(dilation, int):
        dilation = (dilation,)

    return _conv(input_, weight, bias, stride, padding, dilation, groups)


def conv2d(
    input_: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: int | Tuple[int, ...] = 1,
    padding: int | Tuple[int, ...] = 0,
    dilation: int | Tuple[int, ...] = 1,
    groups: int = 1,
) -> Tensor:
    if isinstance(stride, int):
        stride = (stride, stride)
    if isinstance(padding, int):
        padding = (padding, padding)
    if isinstance(dilation, int):
        dilation = (dilation, dilation)

    return _conv(input_, weight, bias, stride, padding, dilation, groups)


def conv3d(
    input_: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: int | Tuple[int, ...] = 1,
    padding: int | Tuple[int, ...] = 0,
    dilation: int | Tuple[int, ...] = 1,
    groups: int = 1,
) -> Tensor:
    if isinstance(stride, int):
        stride = (stride, stride, stride)
    if isinstance(padding, int):
        padding = (padding, padding, padding)
    if isinstance(dilation, int):
        dilation = (dilation, dilation, dilation)

    return _conv(input_, weight, bias, stride, padding, dilation, groups)



from typing import Self, Sequence

from lucid.types import _Scalar, _ArrayOrScalar, _ShapeLike, _ArrayLikeInt


class _TensorOps:
    def __add__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __radd__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __sub__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __rsub__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __mul__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __rmul__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __truediv__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __rtruediv__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __matmul__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __eq__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __ne__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __gt__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __ge__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __lt__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __le__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __pow__(self, _: _Scalar) -> Self: ...

    def __neg__(self) -> Self: ...

    @property
    def T(self) -> Self: ...

    @property
    def mT(self) -> Self: ...

    def dot(self, other: Self | _ArrayOrScalar) -> Self: ...

    def matmul(self, other: Self) -> Self: ...

    def sum(
        self, axis: int | tuple[int] | None = None, keepdims: bool = False
    ) -> Self: ...

    def mean(
        self, axis: int | tuple[int] | None = None, keepdims: bool = False
    ) -> Self: ...

    def var(
        self, axis: int | tuple[int] | None = None, keepdims: bool = False
    ) -> Self: ...

    def reshape(self, *shape: int) -> Self: ...

    def transpose(self, *axes: list[int]) -> Self: ...

    def squeeze(self, axis: _ShapeLike | None = None) -> Self: ...

    def unsqueeze(self, axis: _ShapeLike) -> Self: ...

    def ravel(self) -> Self: ...

    def pad(self, pad_width: _ArrayLikeInt) -> Self: ...

    def repeat(self, repeats: int | Sequence[int], axis: int | None = None) -> Self: ...

    def tile(self, reps: int | Sequence[int]) -> Self: ...

    def flatten(self: Self) -> Self: ...
